// ATtiny Si5351, v1 
// QRP-labs; modified by LZ2WSG
// Selectable frequency version - YO4HFU 23/12/2024
//---------------------------------------------------

#include "TinyWireM.h"

#define SI_I2C           0x62
#define XTAL_FREQ        27000000
#define XTAL_CL          183  
#define OUT_ENABLE_CTRL  3
#define SI_CLK0_CONTROL  16
#define SI_CLK1_CONTROL  17
#define SI_CLK2_CONTROL  18
#define SI_SYNTH_PLL_A   26
#define SI_SYNTH_PLL_B   34
#define SI_SYNTH_MS_0    42
#define SI_SYNTH_MS_1    50
#define SI_SYNTH_MS_2    58
#define SI_PLL_RESET     177
#define SI_R_DIV_1       0b00000000
#define SI_R_DIV_2       0b00010000
#define SI_R_DIV_4       0b00100000
#define SI_R_DIV_8       0b00110000
#define SI_R_DIV_16      0b01000000
#define SI_R_DIV_32      0b01010000
#define SI_R_DIV_64      0b01100000
#define SI_R_DIV_128     0b01110000
#define SI_CLK_SRC_PLL_A 0b00000000
#define SI_CLK_SRC_PLL_B 0b00100000

#define PIN_FREQUENCY_SWITCH 4  // Pin to select frequency
//---------------------------------------------------

uint32_t frequency1 = 2432000;  // First frequency [Hz], switch OPEN
uint32_t frequency2 = 10000000; // Second frequency [Hz], switch CLOSE

//---------------------------------------------------
void setup() {
  TinyWireM.begin();
  pinMode(PIN_FREQUENCY_SWITCH, INPUT_PULLUP);  // Set port 4 as input with pull-up resistor

  si5351_set_frequency(frequency1); // Set initial frequency to frequency1
}
//---------------------------------------------------
void loop() {
  static uint8_t lastPinState = HIGH; // Store the last state of the pin
  uint8_t currentPinState = digitalRead(PIN_FREQUENCY_SWITCH);

  // Check if the pin state has changed (from LOW to HIGH or HIGH to LOW)
  if (currentPinState != lastPinState) {
    // If the pin is HIGH, switch to frequency1, else switch to frequency2
    if (currentPinState == HIGH) {
      si5351_set_frequency(frequency1);  // Switch to frequency1
    } else {
      si5351_set_frequency(frequency2);  // Switch to frequency2
    }
    lastPinState = currentPinState;  // Update the last state of the pin
  }
}
//---------------------------------------------------
void write_reg(uint8_t addr, uint8_t data) {
  TinyWireM.beginTransmission(SI_I2C);
  TinyWireM.send(addr);
  TinyWireM.send(data);
  TinyWireM.endTransmission();
}
//---------------------------------------------------
void setup_pll(uint8_t pll, uint8_t mult, uint32_t num, uint32_t denom) {
  uint32_t p1;
  uint32_t p2;
  uint32_t p3;

  p1 = (uint32_t)(128 * ((float)num / (float)denom));
  p1 = (uint32_t)(128 * (uint32_t)(mult) + p1 - 512);
  p2 = (uint32_t)(128 * ((float)num / (float)denom));
  p2 = (uint32_t)(128 * num - denom * p2);
  p3 = denom;

  write_reg(pll + 0, (p3 & 0x0000FF00) >> 8);
  write_reg(pll + 1, (p3 & 0x000000FF));
  write_reg(pll + 2, (p1 & 0x00030000) >> 16);
  write_reg(pll + 3, (p1 & 0x0000FF00) >> 8);
  write_reg(pll + 4, (p1 & 0x000000FF));
  write_reg(pll + 5, ((p3 & 0x000F0000) >> 12) | ((p2 & 0x000F0000) >> 16));
  write_reg(pll + 6, (p2 & 0x0000FF00) >> 8);
  write_reg(pll + 7, (p2 & 0x000000FF));
}
//---------------------------------------------------
void setup_multisynth(uint8_t synth, uint32_t divider, uint8_t rDiv) {
  uint32_t p1;
  uint32_t p2;
  uint32_t p3;

  p1 = 128 * divider - 512;
  p2 = 0;
  p3 = 1;

  write_reg(synth + 0, (p3 & 0x0000FF00) >> 8);
  write_reg(synth + 1, (p3 & 0x000000FF));
  write_reg(synth + 2, ((p1 & 0x00030000) >> 16) | rDiv);
  write_reg(synth + 3, (p1 & 0x0000FF00) >> 8);
  write_reg(synth + 4, (p1 & 0x000000FF));
  write_reg(synth + 5, ((p3 & 0x000F0000) >> 12) | ((p2 & 0x000F0000) >> 16));
  write_reg(synth + 6, (p2 & 0x0000FF00) >> 8);
  write_reg(synth + 7, (p2 & 0x000000FF));
}
//---------------------------------------------------------------------------------------------------------
void si5351_output_off() {
    write_reg(OUT_ENABLE_CTRL, 0xFF);  // Disable All Outputs
}
//---------------------------------------------------------------------------------------------------------
void si5351_output_on() {
    write_reg(OUT_ENABLE_CTRL, 0xFE);  // Enable Output CLK_0
    }
//---------------------------------------------------------------------------------------------------------
void si5351_XTAL_cap_load() {
  write_reg(XTAL_CL, 0xD2); // 0x12 - 0pF / 0x52 - 6pF / 0x92 - 8pF / 0xD2 - 10pF (default) - TCXO injection is not affected
} 
//---------------------------------------------------------------------------------------------------------
  void si5351_PLL_RESET(){
  write_reg(SI_PLL_RESET, 0xA0);   // reset PLL A 
  }
//---------------------------------------------------------------------------------------------------------
  void si5351_CLK_0_CONTROL(){
  write_reg(SI_CLK0_CONTROL, 0x4F);  // Power up CLK0
}
//---------------------------------------------------------------------------------------------------------
void si5351_CLK_1_CONTROL(){
  write_reg(SI_CLK1_CONTROL, 0x80);  // Power down CLK1
}
//---------------------------------------------------------------------------------------------------------
void si5351_CLK_2_CONTROL(){
  write_reg(SI_CLK2_CONTROL, 0x80);  // Power down CLK2
}
//---------------------------------------------------------------------------------------------------------
void si5351_set_frequency(uint32_t frequency) {
  uint32_t pllFreq;
  uint32_t xtalFreq = XTAL_FREQ;
  uint32_t l;
  float f;
  uint8_t mult;
  uint32_t num;
  uint32_t denom;
  uint32_t divider;

 divider = 900000000 / frequency;
  if (divider % 2)
    divider--;

  pllFreq = divider * frequency;

  mult = pllFreq / xtalFreq;
  l = pllFreq % xtalFreq;
  f = l;
  f *= 1048575;
  f /= xtalFreq;
  num = f;
  denom = 1048575;

 // si5351_output_off();  // outputs disable not used
  
  si5351_XTAL_cap_load(); 
     
  setup_pll(SI_SYNTH_PLL_A, mult, num, denom);
  
  setup_multisynth(SI_SYNTH_MS_0, divider, SI_R_DIV_1); 
   
  si5351_PLL_RESET();   

 // si5351_output_on();   // enable output CLK_0, not used. Outputs selected by power-up / power-down
 
  si5351_CLK_0_CONTROL(); // power-up CLK0
  
  si5351_CLK_1_CONTROL(); // power-down CLK1
  
  si5351_CLK_2_CONTROL(); // power-down CLK2  
}
